# ------------------------------------------------------------
# Auto generated test file `test_054_SharedVar.py`
# ------------------------------------------------------------
# source file   = NBTest_054_SharedVar.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 054
# test comment  = SharedVar
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonOrderUI, P, __version__, __date__
from carbon.helpers import SharedVar, print_version
from math import floor, ceil, trunc
print_version(require="2.2.4")




# ------------------------------------------------------------
# Test      054
# File      test_054_SharedVar.py
# Segment   Shared Var
# ------------------------------------------------------------
def test_shared_var():
# ------------------------------------------------------------
    
    # #### Definitions
    
    x, x2 = SharedVar(5), SharedVar(5.1234)
    x
    
    assert repr(x)[:17] == "SharedVar(5, oid="
    
    # #### Comparisons
    
    assert (x==x, x==5, x<5, x>5, x<=5, x>=5, 5<x, 5>x) == (True, True, False, False, True, True, False, False)
    assert (x==x, 5<=x, 5>=x) == (True, True, True)
    
    # #### Unitary operations
    
    assert x==+x
    assert x.id != (+x).id
    assert (-x).value == -5
    assert x.id != (-x).id
    assert (abs(-x)).value == 5
    assert x.id != (abs(-x)).id
    assert round(x2,2)==5.12
    assert x2.id != round(x2,2).id
    assert trunc(x2)==5
    assert x2.id != trunc(x2).id
    assert floor(x2)==5
    assert x2.id != floor(x2).id
    assert ceil(x2)==6
    assert x2.id != ceil(x2).id
    
    # #### Binary operations
    
    assert x+1 == 6
    assert x.id != (x+1).id
    assert x-1 == 4
    assert x.id != (x-1).id
    assert x*2 == 10
    assert x.id != (x*2).id
    assert x/2 == 2.5
    assert x.id != (x/2).id
    assert x//2 == 2
    assert x.id != (x//2).id
    assert x**2 == 25
    assert x.id != (x//2).id
    assert x<<1 == 10
    assert x.id != (x<<1).id
    assert x>>1 == 2
    assert x.id != (x>>1).id
    assert x&3 == 1
    assert x.id != (x&3).id
    assert x|2 == 7
    assert x.id != (x|2).id
    assert x^1 == 4
    assert x.id != (x^1).id
    
    assert divmod(x,2)[0] == 2
    assert divmod(x,2)[1] == 1
    assert x.id != divmod(x,2)[0].id
    assert x.id != divmod(x,2)[1].id
    
    # #### Reverse binary operations
    
    assert 1+x == 6
    assert x.id != (1+x).id
    assert 1-x == -4
    assert x.id != (1-x).id
    assert 2*x == 10
    assert x.id != (2*x).id
    assert 2/x == 0.4
    assert x.id != (2/x).id
    assert 2//x == 0
    assert x.id != (2//x).id
    assert 2**x == 32
    assert x.id != (2**x).id
    assert 1<<x == 32
    assert x.id != (1<<x).id
    assert 1024>>x == 32
    assert x.id != (1024>>x).id
    assert 3&x == 1
    assert x.id != (3&x).id
    assert 2|x == 7
    assert x.id != (2|x).id
    assert 1^x == 4
    assert x.id != (1^x).id
    
    assert divmod(16,x)[0] == 3
    assert divmod(16,x)[1] == 1
    assert x.id != divmod(16,x)[0].id
    assert x.id != divmod(16,x)[1].id
    
    # #### Assignment operations
    
    y = SharedVar(100)
    id0 = y.id
    y+=1
    y-=1
    y*=10
    y//=10
    assert y == 100
    y<<=2
    y>>=1
    assert y == 200
    y&=199
    y|=1
    y^=17
    assert y == 208
    y /= 32
    assert y == 6.5
    assert y.id == id0
    
    # #### Conversions
    
    x = SharedVar(5.5)
    assert int(x) == 5
    assert str(x) == "5.5"
    assert float(x) == 5.5
    assert complex(x) == complex(float(x))
    
    
    
    
    


"""
Carbon -- data generation library
"""
__VERSION__ = "1.0"
__DATE__ = "10/01/2023"

from math import sqrt, exp, log
from datetime import timedelta as _timedelta, datetime as _datetime
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

class PathGenerator():
    """
    generate random lognormal paths
    
    :mu:       drift parameter; eg 0.05 = 5%
    :sig:      volatility parameter; eg 0.2 = 20%
    :s0:       spot value at time zero
    :time:     time period (in years)
    :N:        number of steps (excluding t=0)
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__
    
    DEFAULTS = {
        "mu": 0,
        "sig": 0.5,
        "s0": 100,
        "time": 1,
        "N": 200,
        "startdt": (2020, 1, 1),
        "numpaths": 100,
        "colname": "p{sig100:03d}-{mu100:02d}-{ix:03d}",
    }
    
    def __init__(self, mu=None, sig=None, s0=None, time=None, N=None, startdt=None, colname=None):
        
        if mu is None: 
            mu = self.DEFAULTS["mu"]
        if sig is None: 
            sig = self.DEFAULTS["sig"]
        if s0 is None: 
            s0 = self.DEFAULTS["s0"]        
        if time is None: 
            time = self.DEFAULTS["time"] 
        if N is None: 
            N = self.DEFAULTS["N"]  
        if startdt is None: 
            startdt = self.DEFAULTS["startdt"]  
        if colname is None:
            colname = self.DEFAULTS["colname"] 
            
        self.mu = mu
        self.sig =sig
        self.s0 = s0
        self.time = time
        self.N = N
        self.startdt = startdt
        self.colname = colname
        
        self.dt = time/N
        self.dt_td = _timedelta(days=self.dt*365.25)
        self.mudt = self.mu*self.dt
        self.halfsig2dt = 0.5*self.sig*self.sig*self.dt
        self.volsqrtdt = self.sig*sqrt(self.dt)
        self.times = np.array([i*self.dt for i in range(self.N+1)])
        self.datetimes = self._datetime_r(self.startdt)
            

    def path(self, plot=False):
        """
        generates a new random path
        
        :plot:     if True, plot the path
        """
        increments = np.random.default_rng().lognormal(
                        mean=        self.mudt - self.halfsig2dt, 
                        sigma=       self.volsqrtdt, 
                        size=        self.N)
        
        path = np.cumprod(np.insert(increments, 0, self.s0))
        if plot:
            self._plot(path)
        return path
    
    def _plot(self, path=None, datetimes=None):
        """
        plots the path with maplotlib
        """
        if path is None: path = self.path()
        if datetimes is None: datetimes = self.datetimes
        
        plt.title(f"random path (sig={self.sig*100:.0f}%, mu={self.mu*100:.0f}%, N={self.N})")
        plt.xlabel("time")
        plt.ylabel("spot")
        plt.plot(datetimes, path)
        plt.grid()
    
    def _datetime_r(self, startdt=None):
        """
        generates a datetime range
        
        :startdt:    either a datetime object, or tuple (day, month, year)
        """
        if startdt is None: 
            startdt = self.startdt
        if not isinstance(startdt, _datetime):
            if not len(startdt) == 3:
                raise ValueError("startdate must be a tuple of lenght 3", startdt)
            startdt = _datetime(*startdt)
        return np.array([startdt+i*self.dt_td for i in range(self.N+1)])
    
    @staticmethod
    def vec2ar(paths):
        """concatenates list of vectors into a single numpy array"""
        return np.concatenate([p.reshape(-1,1) for p in paths], axis=1)
    
    def pathdf(self, numpaths=None, colname=None):
        """
        generate a pandas dataframe with newly generated paths
        
        :numpaths:    number of paths to generate
        :colname:     column name (format string; {i} is the path index)
        """
        
        if numpaths is None: numpaths = self.DEFAULTS["numpaths"]
        if colname is None: colname = self.DEFAULTS["colname"]
        
        paths = [self.path() for _ in range(numpaths)]
        #df0 = pd.DataFrame(data=self.vec2ar([self.times, self.datetimes]), columns=["time", "datetime"])
        df0 = pd.DataFrame(data=self.vec2ar([self.datetimes]), columns=["datetime"])
        cols = [
            colname.format(
                sig100 = int(self.sig*100),
                mu100  = int(self.mu*100),   
                ix     = ix
                ) 
            for ix in range(len(paths))
        ]
        df1 = pd.DataFrame(data=self.vec2ar(paths), columns=cols)
        df = pd.concat([df0, df1], axis=1).set_index("datetime")
        return df
    
    def __call__(self):
        """
        alias for path
        """
        return self.path()



"""
Carbon helper module -- check version requirements are met

    require_version()
"""
import re as _re
from .. import __version__

class VersionRequirementNotMetError(RuntimeError): pass

def split_version_str(vstr):
    """splits version mumber string into tuple (int, int, int, ...)"""
    m = _re.match("^([0-9\.]*)", vstr.strip())
    if m is None:
        raise ValueError("Invalid version number string", vstr)
    vlst = [int(x) for x in m.group(0).split(".")]
    return vlst

def require_version(required, actual=None, raiseonfail=True):
    """
    checks whether required version is smaller or equal actual version
    
    :required:    the required version, eg "1.2.1"
    :actual:      the actual version, eg "1.3-beta2"
    :returns:     True if requirements are met, False else*
    
    *note: "1.3-beta1" MEETS the requirement "1.3"!
    """
    if actual is None: actual = __version__

    rl, al = split_version_str(required), split_version_str(actual)
    #print(f"required={rl}, actual={al}")

    result = _require_version(rl,al)
    if not raiseonfail:
        return result
    if not result:
        raise VersionRequirementNotMetError(f"Version requirements not met (required = {rl}, actual = {al})", required, actual)

def _require_version(rl, al):
    """
    checks whether required version is smaller or equal actual version
    
    :rl:        the required version eg (1,2,1)
    :al:        the actual version, eg (1,3)
    :returns:   True if requirements are met, False else*
    """
    for r,a in zip(rl, al):
        #print(f"r={r}, a={a}")
        if r > a:
            return False
        elif r < a:
            return True
    return len(al) >= len(rl)
    


"""
Carbon helper module - retrieve data from CryptoCompare
"""
__VERSION__ = "1.1"
__DATE__ = "27/Jan/2023"

import os as _os
import pandas as _pd
import hashlib as _hashlib
import requests as _requests
import pickle as _pickle
from collections import namedtuple as _namedtuple


pair_t = _namedtuple("pair", "tknb,tknq")

class CryptoCompare():
    """
    simple class formalizing interaction with the crypto compare API 
    
    :apikeyname:    the OS environment variable holding the API key
                    only used if no `apikey`; default is class.APIKEYNAME
    :apikey:        the API key; if True use without API key
    :datapath:      the path where all data is written (and read from)
    :raiseonerror:  if True, errors usually lead to an exception, otherwise to a None return
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    BASEURL = "https://min-api.cryptocompare.com" # must NOT end with /
    APIKEYNAME = "CCAPIKEY" # the name of the environment variable containing the API key
    RAISEONERROR = True
    DATAPATH = "cryptocompare"
    
    DEFAULT_TSYM = "usd"
    DEFAULT_LIMIT = 2000
    
    def __init__(self, apikeyname=None, apikey=None, raiseonerror=None):
        if raiseonerror is None:
            raiseonerror = self.RAISEONERROR
        self.raiseonerror = raiseonerror        
        if apikey is None:
            if apikeyname is None:
                apikeyname = self.APIKEYNAME
            apikey = _os.getenv(self.keyname)
            if apikey is None:
                print(f"Can't find API key {apikeyname} in environment variables.")
                print(f"Use `export {apikeyname}=<value>` to set it BEFORE you launch Jupyter")
                raise RuntimeError(f"API key not present. Use `export {apikeyname}=<value>` to set it before launching Jupyter.")
        self.apikey = apikey
    
    def url(self, endpoint):
        """
        returns the URL of a given endpoint
        """
        return f"{self.BASEURL}{endpoint}"
    
    @property
    def keydigest(self):
        """returns signature (=SHA1 hash) of the API key, or 0000... if anonymous"""
        if self.apikey is True:
            return "0"*40
        return _hashlib.sha1(self.apikey.encode()).hexdigest()
    
    def datafn(self, fn):
        """returns the full data file name, including path"""
        return _os.path.join(self.DATAPATH, fn)

    def cache(self, item):
        """
        reads a data item from the data cache
        """
        try:
            with open(self.datafn(f"{item}.pickle"), "rb") as f:
                result = _pickle.load(f)
        except:
            if not self.raiseonerror:
                return None
            raise
        return result
    
    def write_cache(self, item, data):
        """
        writes `data` to the cache under the name `item`
        
        :returns:    `item` on success, None (or raises) on failure
        """
        try:
            with open(self.datafn(f"{item}.pickle"), "wb") as f:
                _pickle.dump(data, f)
        except:
            if not self.raiseonerror:
                return None
            raise           
        return item
    
    QUERY_GET = "GET"
    QUERY_POST = "POST"
    def query(self, endpoint, params=None, method=None):
        """
        generic API query
        
        :endpoint:  the API endpoint to call, eg "/all/exchanges"
        :params:    the API parameters (parameters with value None will be removed)
        :method:    http method; default is QUERY_GET
        """
        if method is None:
            method = self.QUERY_GET
        if params is None:
            params = dict()
        url = self.url(endpoint)
        paramsq = {k:v for k,v in params.items() if not v is None}
        print("[query]", url, paramsq, f"[{str(self.keydigest)[:4]}]")
        if not self.apikey is True:
            paramsq["api_key"] = self.apikey
        
        if method == self.QUERY_GET:
            r = _requests.get(url, params=paramsq)
        elif method == self.QUERY_POST:
            raise ValueError("Method QUERY_POST has not been implemented yet.")
        else:
            raise ValueError("Unknown method. Use QUERY_XXX constants", method)
        
        if not r:
            if self.raiseonerror:
                raise RuntimeError(f"API query not successfull (status={r.status})", r)
            else:
                return None
        return r
    
    def query_allexchanges(self):
        """
        endpoint = /data/v4/all/exchanges
        
        https://min-api.cryptocompare.com/documentation?key=Other&cat=allExchangesV4Endpoint
        """
        r = self.query(
            endpoint="/data/v4/all/exchanges"
        )
        if r is None: return r
        return r.json().get("Data")

    
    def _cache_xxx(self, item, updatemethod, readonfail=True, updateonfail=False):
        """
        generic cached access
        
        :item:           the name of the item in the cache
        :updatemethod:   the method to call for updating it
        :readonfail:     if True, on cache miss updatemethod is called
        :updateonfail:   it True, on cache miss, updatemethod is called an item is 
                         written to cache
        """
        if updateonfail:
            readonfail = True
        try:
            return self.cache(item)
        except:
            print(f"[_cache_xxx] cache miss for item {item}")
            if readonfail:
                print(f"[_cache_xxx] reading {item} from API")
                data = updatemethod()
                if updateonfail:
                    print(f"[_cache_xxx] updating cache for {item} from API")
                    self.write_cache(item, data)
                return data
            else:
                if self.raiseonerror:
                    raise
                else:
                    return None
        
    def cache_allexchanges(self, readonfail=True, updateonfail=False):
        """cached access to query_allexchanges"""
        return self._cache_xxx(
            item="query_allexchanges", 
            updatemethod=self.query_allexchanges
        )
    
    def query_ratelimit(self):
        """
        endpoint = /stats/rate/limit
        
        https://min-api.cryptocompare.com/documentation?key=Other&cat=rateLimitEndpoint
        """
        r = self.query(
            endpoint="/stats/rate/limit"
        )
        if r is None: return r
        return r.json().get("Data")
        
    def query_coinlist(self):
        """
        endpoint = /data/all/coinlist
        
        https://min-api.cryptocompare.com/documentation?key=Other&cat=allCoinsWithContentEndpoint
        """
        r = self.query(
            endpoint="/data/all/coinlist"
        )
        if r is None: return r
        return r.json().get("Data")
    
    def cache_coinlist(self, readonfail=True, updateonfail=False):
        """cached access to query_coinlist"""
        return self._cache_xxx(
            item="query_coinlist", 
            updatemethod=self.query_coinlist
        )
        
    def query_indexlist(self):
        """
        endpoint = /data/index/list
        
        https://min-api.cryptocompare.com/documentation?key=Index&cat=listOfIndices
        """
        r = self.query(
            endpoint="/data/index/list"
        )
        if r is None: return r
        return r.json().get("Data")
    
    def cache_indexlist(self, readonfail=True, updateonfail=False):
        """cached access to query_indexlist"""
        return self._cache_xxx(
            item="query_indexlist", 
            updatemethod=self.query_indexlist
        )
    
    @staticmethod
    def ts_tocc(ts):
        """
        convert timestamp into format needed by CryptoCompare
        
        :ts:     the timestamp in any format that works for pd.Timestamp(ts)
        """
        return int(_pd.Timestamp(ts).timestamp())
    
    @staticmethod
    def ts_fromcc(ts):
        """
        convert timestamp from CryptoCompare format into pd.Timestamp format
        """
        return _pd.to_datetime(ts, unit='s', origin='unix') 
    
    def query_dailypair(self, fsym=None, tsym=None, e=None, limit=False, toTs=None, aspandas=True):
        """
        endpoint = /data/v2/histoday
        
        :fsym:     cryptocurrency symbol of interest
        :tsym:     currency symbol to convert into
        :e:        exchange to obtain data from
        :limit:    number of data points to return (max: 2000; False defaults to that number)
        :toTs:     returns historical data BEFORE that timestamp
                   timestamp format either 1452680400 or pd.Timestamp compatible string 
        
        https://min-api.cryptocompare.com/documentation?key=Historical&cat=dataHistoday
        """
        r = self.query(
            endpoint="/data/v2/histoday",
            params = {
                "fsym":     fsym,
                "tsym":     tsym if not tsym is None else self.DEFAULT_TSYM,
                "e":        e,
                "limit":    limit if not limit is False else self.DEFAULT_LIMIT,
                "toTs":     toTs,
            }
        )
        if r is None: return r
        if not aspandas:
            return r.json().get("Data")
        try:   
            df = _pd.DataFrame.from_records(r.json()["Data"]["Data"])
            df["datetime"] = [self.ts_fromcc(ts) for ts in df["time"]]
            df = df.set_index("datetime")
            del df["conversionType"]
            del df["conversionSymbol"]
            del df["time"]
            df = df[['open', 'close', 'high', 'low', 'volumefrom', 'volumeto']]
            return df
        except:
            if self.raiseonerror: raise
            return None

    def ccycodes(self, symonly=True, fn=None):
        """
        returns information on currency codes
        
        :symonly:   if True (default) only return list of ccy symbold
        :fn:        the filename of the currency code file
        """
        if symonly:
            return self.join( self.unjoin(self.CCYCODES) )
        if fn is None:
            fn = _os.path.join(self.DATAPATH, "isoccy.csv")
        df = _pd.read_csv(fn, index_col=False)
        if symonly:
            symbols = list(set(df["Symbol"]))
            symbols.sort()
            return tuple(symbols)
        return df
    
    CCYCODES = """
    AED,AFN,ALL,AMD,ANG,AOA,ARS,AUD,AWG,AZN,BAM,BBD,BDT,BGN,BHD,BIF,BMD,
    BND,BOB,BOV,BRL,BSD,BTN,BWP,BYN,BZD,CAD,CDF,CHE,CHF,CHW,CLF,CLP,CNY,
    COP,COU,CRC,CUC,CUP,CVE,CZK,DJF,DKK,DOP,DZD,EGP,ERN,ETB,EUR,FJD,FKP,
    GBP,GEL,GHS,GIP,GMD,GNF,GTQ,GYD,HKD,HNL,HRK,HTG,HUF,IDR,ILS,INR,IQD,
    IRR,ISK,JMD,JOD,JPY,KES,KGS,KHR,KMF,KPW,KRW,KWD,KYD,KZT,LAK,LBP,LKR,
    LRD,LSL,LYD,MAD,MDL,MGA,MKD,MMK,MNT,MOP,MRU,MUR,MVR,MWK,MXN,MXV,MYR,
    MZN,NAD,NGN,NIO,NOK,NPR,NZD,OMR,PAB,PEN,PGK,PHP,PKR,PLN,PYG,QAR,RON,
    RSD,RUB,RWF,SAR,SBD,SCR,SDG,SEK,SGD,SHP,SLL,SOS,SRD,SSP,STN,SVC,SYP,
    SZL,THB,TJS,TMT,TND,TOP,TRY,TTD,TWD,TZS,UAH,UGX,USD,USN,UYI,UYU,UYW,
    UZS,VES,VND,VUV,WST,XAF,XAG,XAU,XCD,XDR,XOF,XPD,XPF,XPT,XSU,XUA,YER,
    ZAR,ZMW,ZWL
    """.strip()

    @staticmethod
    def join(tpl, sep=None):
        """join the tpl into comma separated strings"""
        if sep is None: sep = ", "
        return sep.join(str(s) for s in tpl)
        
    @staticmethod
    def unjoin(jstr, filter=None, sep=None):
        """
        unjoin the join string, stripping the result
        
        :jstr:      a (typically comma) separated string
        :filter:    filter to be applied (default: str)
        :sep:       the separator (default: comma)
        :returns:   tuple
        """
        if sep is None: sep = ","
        result = jstr.split(sep)
        if filter is None:
            filter = str
        result = ( filter(c.strip()) for c in result)
        return tuple(result)

    @staticmethod
    def pairs_fields_from_df(df):
        """
        pairs and fields present in the dataframe
        
        :df:        data frame with index = (base token, quote token, field)
        :returns:   dict pairs: tuple( (tknp1, tnkq1), ...), fields: (field1, ...)
        """
        pairs  = ((tknb, tknq) for tknb, tknq, field in df.columns)
        pairs  = tuple(set(pairs))
        fields = (field for tknb, tknq, field in df.columns)
        fields = tuple(set(fields))
        return {"pairs": pairs, "fields": fields}

    DEFAULT_DF_FIELD = "close"
    def reformat_raw_df(self, df, field=None):
        """
        reformats a raw df
        
        :df:     the raw df, as returned by a concatenation eg of daily_pair calls
        :field:  the name of the price field to use for the price (default: DEFAULT_DF_FIELD)
        """
        if field is None:
            field = self.DEFAULT_DF_FIELD
        return _pd.concat(
            [
                df[(*pair, "open")].rename(f"{pair[0]}/{pair[1]}", inplace=True)
                for pair in self.pairs_fields_from_df(df)["pairs"]
            ], axis=1
        )

    @staticmethod
    def pt_from_pair(pair):
        """
        creates a pair tuple (tknb, tknq) from a pair 'TKNB/TKNQ'
        """
        return pair_t(*pair.split("/"))
    
    @staticmethod
    def pair_from_pt(pair_t):
        """
        creates a pair 'TKNB/TKNQ' from a pair tuple (tknb, tknq)
        """
        return "/".join(pair_t)

    @staticmethod
    def coinlist(coins, sep=","):
        """
        creates a coin list from separated string (does not touch lists)
        """
        if isinstance(coins, str):
            return tuple(c.strip() for c in coins.split(sep))
        else:
            return coins

    @classmethod
    def create_pairs(cls, coins, quotecoins=None):
        """
        create pair tuples from all possible combinations of coins and quotecoins
        
        :coins:        a list of coins, either ("tkn1", "tkn2") or "tkn1, tkn2"
        :quotecoins:   a list of quote coins; if None set equal to coins
        :returns:      all combinations as tuples (c, qc) with c!=qc
        """
        coins = cls.coinlist(coins)
        if quotecoins is None:
            quotecoins = coins
        else:
            quotecoins = cls.coinlist(quotecoins)
        result = ( (c,q) for q in quotecoins for c in coins)
        result = ( pair_t(c,q) for c,q in result if c != q)
        return tuple (result)

      


"""
Carbon helper module -- useful ipywidgets
"""
__VERSION__ = "1.0"
__DATE__ = "23/01/2023"

import ipywidgets as _ipw

class CheckboxManager():
    """
    manages checkboxes that create True/False dicts
    
    :choices:       the checkbox choices either as list/tuple, or as dict; if it is a dict, then
                    the keys of the dict are the IDs, and the values are the display text; if it
                    is a list/tuple then keys and display values are the same
    :values:        the initial values; either None, False (unchecked), True (checked), or a
                    tuple/list of bools, one for each of the boxes
    :objects:       the objects associated with the checkbox items; if given, the `objects` method
                    returns all (un)checked objects
    :disabled:      if None, all False; otherwise tuple/list of bools
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    def __init__(self, choices, values=None, objects=None, disabled=None):
    
        self.choice_ids = tuple(str(x) for x in choices) 
        try:
            self.choice_descr = tuple(str(x) for x in choices.values())
        except AttributeError:
            self.choice_descr = self.choice_ids
        
        if objects is None:
            self._objects = self.choice_descr
        else:
            if not len(objects) == len(self.choice_ids):
                raise ValueError(f"len(objects) [{len(objects)}] must be len(choices) [{len(self.choice_ids)}]", objects, choices)
            self._objects = objects
        if values is None or values is False:
            values = (False,)*len(self.choice_ids)
        elif values is True:
            values = (True,)*len(self.choice_ids)
        elif len(values) == len(self.choice_ids):
            values = tuple(bool(v) for v in values)
        else:
            raise ValueError(f"len(values) [{len(values)}] must be len(choices) [{len(self.choice_ids)}]", 
                             values, choices)
        self.initial_values = tuple(values)
        if disabled is None:
            self.disabled = (False,)*len(self.choice_ids)
        else:
            self.disabled = tuple(bool(v) for v in disabled)
            if len(self.disabled) != len(self.choice_ids):
                raise ValueError(f"len(disabled) [{len(self.disabled)}] must be len(choices) [{len(self.choice_ids)}]", 
                                 disabled, choices)
            
        self._checkboxes = tuple(
            _ipw.Checkbox(
                value=val, 
                description=descr, 
                disabled=dis,
                #layout = _ipw.Layout(width="150px")
            )
            for val, descr, dis in zip(self.initial_values, self.choice_descr, self.disabled)
        )
        
    @classmethod
    def from_idvdct(cls, idvdct, disabled=None):
        """
        alternative constructor, using idvdict
        
        :idvdct:     the id/value dict, ie a dict id => value
        :disabled:   as in the main constructor 
        """
        return cls(choices=idvdct.keys(), values=idvdct.values(), disabled=disabled)
    
    @property
    def checkboxes(self):
        """the tuple of check boxes"""
        return self._checkboxes
    
    @property
    def vbox(self):
        """the widget box, as vbox"""
        return _ipw.VBox(self.checkboxes)
        # https://ipywidgets.readthedocs.io/en/7.6.3/examples/Widget%20Styling.html#The-Grid-layout
        # https://ipywidgets.readthedocs.io/en/7.6.3/examples/Widget%20Styling.html#The-Flexbox-layout

    @property
    def hbox(self):
        """the widget box, as vbox"""
        return _ipw.HBox(self.checkboxes)   
    
    @property
    def values_dct(self):
        """returns a dict choice_id => chosen value"""
        return {cid: cb.value for cid, cb in zip(self.choice_ids, self._checkboxes)}
    
    @property
    def values(self):
        """returns tuple of chosens values"""
        return tuple(cb.value for cb in self._checkboxes)

    def display(self, vertical=True):
        """short for display(self.box)"""
        display(self.vbox if vertical else self.hbox)

    @property
    def checked(self):
        """returns a tuple of the ids of all checked boxes"""
        return tuple(cid for cid, val in self.values_dct.items() if val == True)
    
    @property
    def unchecked(self):
        """returns a tuple of the ids of all unchecked boxes"""
        return tuple(cid for cid, val in self.values_dct.items() if val == False)
    
    def objects(self, checked=True):
        """returns a tuple of objects pertaining to the (un)checked boxes"""
        return tuple(obj for obj, cb in zip(self._objects, self._checkboxes) if cb.value == checked)
    
    def __call__(self, *args, **kwargs):
        self.display(*args, **kwargs)

class PcSliderManager():
    """
    manages sliders that display percentages (0..100)
    
    :choices:       the slider choices either as list/tuple, or as dict; if it is a dict, then
                    the keys of the dict are the IDs, and the values are the display text; if it
                    is a list/tuple then keys and display values are the same
    :values:        the initial values; must be between 0..1 [-1..1 for pm sliders], even though 
                    the percentages are displayed as between 0..100%
    :pm:            plusminus; if None or False, range is 0..100%; if True, -100%..100%
    :range:         either a tuple (min, max), or a range of tuples; must not be provided at the 
                    same time as pm
    :disabled:      if None, all False; otherwise tuple/list of bools
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    def __init__(self, choices, values=None, pm=None, range=None, disabled=None):
    
        if not range is None and not pm is None:
            raise ValueError("pm and range must not be not-None at the same time")

        self.choice_ids = tuple(str(x) for x in choices) 
        try:
            self.choice_descr = tuple(str(x) for x in choices.values())
        except AttributeError:
            self.choice_descr = self.choice_ids
            
        if values is None:
            values = (0.,)*len(self.choice_ids)
        else:
            try:
                if values >= 0 and values <= 1:
                    values = (float(values),)*len(self.choice_ids)
                else:
                    raise ValueError("Value must be between 0..1", values)
            except TypeError:
                if len(values) == len(self.choice_ids):
                    values = tuple(float(v) for v in values)
                else:
                    raise ValueError(f"len(values) [{len(values)}] must be len(choices) [{len(self.choice_ids)}]", values, choices)         
        self.initial_values = tuple(values)
        
        if disabled is None:
            self.disabled = (False,)*len(self.choice_ids)
        else:
            self.disabled = tuple(bool(v) for v in disabled)
            if len(self.disabled) != len(self.choice_ids):
                raise ValueError(f"len(disabled) [{len(self.disabled)}] must be len(choices) [{len(self.choice_ids)}]", disabled, choices)
        
        if pm is None or pm is False:
            self.pm = (False,)*len(self.choice_ids)
        elif pm is True:
            self.pm = (True,)*len(self.choice_ids)
        else:
            self.pm = tuple(bool(v) for v in pm)
            if len(self.pm) != len(self.choice_ids):
                raise ValueError(f"len(pm) [{len(self.disabled)}] must be len(choices) [{len(self.choice_ids)}]", disabled, choices)
        
        if not range is None:
            try:
                range[0][0]
                # this must be a tuple of tuples
                self.range = tuple(tuple(v) for v in range)
                if len(self.range) != len(self.choice_ids):
                    raise ValueError(f"len(range) [{len(self.range)}] must be len(choices) [{len(self.choice_ids)}]", range, choices)
            except:
                self.range = (range,)*len(self.choice_ids)
        else:
            self.range = ( (-1, 1) if pmv else (0,1) for pmv in self.pm)

        self._sliders = tuple(
            _ipw.FloatSlider(
                value=val*100,
                min=rg[0]*100,
                max=rg[1]*100,
                step=0.1,
                description=descr,
                disabled=dis,
                continuous_update=False,
                orientation='horizontal',
                readout=True,
                readout_format='.1f',
            )
            for val, descr, dis, rg in zip(
                self.initial_values, 
                self.choice_descr, 
                self.disabled,
                self.range,
            )
        )
        
#     @classmethod
#     def from_idvdct(cls, idvdct, disabled=None):
#         """
#         alternative constructor, using idvdict
        
#         :idvdct:     the id/value dict, ie a dict id => value
#         :disabled:   as in the main constructor 
#         """
#         return cls(choices=idvdct.keys(), values=idvdct.values(), disabled=disabled)
    
    @property
    def sliders(self):
        """the tuple of check boxes"""
        return self._sliders
    
    @property
    def vbox(self):
        """the widget box, as vbox"""
        return _ipw.VBox(self.sliders)
        # https://ipywidgets.readthedocs.io/en/7.6.3/examples/Widget%20Styling.html#The-Grid-layout
        # https://ipywidgets.readthedocs.io/en/7.6.3/examples/Widget%20Styling.html#The-Flexbox-layout

    @property
    def hbox(self):
        """the widget box, as vbox"""
        return _ipw.HBox(self.sliders)   
    
    @property
    def values_dct(self):
        """returns a dict choice_id => chosen value"""
        return {cid: cb.value/100. for cid, cb in zip(self.choice_ids, self._sliders)}
    
    @property
    def values(self):
        """returns tuple of chosens values"""
        return tuple(cb.value/100. for cb in self._sliders)

    def display(self, vertical=True):
        """short for display(self.box)"""
        display(self.vbox if vertical else self.hbox)
        
    def __call__(self, *args, **kwargs):
        self.display(*args, **kwargs)


class DropdownManager():
    """
    manages dropdowns that create True/False dicts
    
    :options:       tuple/list of dropdown options; alternatively a dict, in which case the 
                    dict keys are the options, and the values are the labels
    :labels:        the labels that are displayed in the dropdown
    :descr:         an option string displayed in front of the dropdown
    :defaultix:     index of the default option chosen
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    def __init__(self, options, labels=None, defaultval=None, defaultix=None, descr=None):
    
        if not defaultval is None and not  defaultix is None:
            raise ValueError("Only one of defaultval, defaultix can be None")

        self.options = tuple(str(x) for x in options) 
        try:
            self.labels = tuple(str(x) for x in options.values())
            if not labels is None:
                raise ValueError("Must not provide labels in option dict and explicitly", options, labels)
        except AttributeError:
            if labels is None:
                self.labels = self.options
            else:
                self.labels = tuple(str(x) for x in labels)
        
        self.options = tuple(str(x) for x in options) 

        if not descr: 
            descr = ""
        self.descr = str(descr)

        if not defaultval is None:
            self.defaultix = None
            self.defaultval = defaultval
            self._dropdown = _ipw.Dropdown(
                options=[(l,o) for l,o in zip(self.labels, self.options)],
                value=self.defaultval,
                description=self.descr,
                disabled=False,
            )
        else:
            if defaultix is None:
                defaultix = 0
            if not defaultix >=0 and defaultix < len(self.options):
                raise ValueError(f"must have 0 <= defaultix [{defaultix}] < len(options) [{len(self.options)}]")
            self.defaultix = defaultix
            self.defaultval = None
            self._dropdown = _ipw.Dropdown(
                options=[(l,o) for l,o in zip(self.labels, self.options)],
                index=self.defaultix,
                description=self.descr,
                disabled=False,
            )

    @property
    def dropdown(self):
        """returns the _dropdown"""
        return self._dropdown
    
    @property
    def value(self):
        """returns the dropdown value"""
        return self._dropdown.value
    
    @property
    def label(self):
        """returns the dropdown label"""
        return self._dropdown.label

    @property
    def result(self):
        """returns the dropdown value and label"""
        return self._dropdown.value, self._dropdown.label
    
    def display(self):
        """short for display(self.dropdown)"""
        display(self.dropdown)
        
    def __call__(self):
        self.display()



"""
Carbon helper module - standard Carbon imports for sims

USAGE

    from carbon.helper.stdimports import *

This will print version numbers of key Carbon modules
"""
__VERSION__ = "1.1"
__DATE__ = "25/01/2023"

from math import sqrt, exp, log
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import os

print("[stdimports] imported np, pd, plt, os, sqrt, exp, log")

from .. import CarbonSimulatorUI, CarbonOrderUI, analytics as cal, P, __version__, __date__
from .version import require_version
from . import print_version, plt_style




"""
Carbon helper module - sundry helpers
"""
__VERSION__ = "1.0"
__DATE__ = "23/01/2023"

import os as _os

def listdir(path, ext=None):
    """
    get a sorted directory listing, filtered by extension; removes extension
    
    :path:     the path of the directory to be listed
    :ext:      the extension (eg ".pickle")
    :returns:  sorted list of filenames, ext removed
    """
    if ext is None:
        flist = _os.listdir(path)
    else:
        ext = ext.lower()
        l = len(ext)
        flist = [fn[:-l] for fn in _os.listdir(path) if fn[-l:].lower() == ext]
    flist.sort()
    return flist


"""
Carbon helper module - run the simulation
"""
__VERSION__ = "3.0"
__DATE__ = "28/01/2023"

from collections import namedtuple
import numpy as _np
import pandas as _pd
from math import sqrt
from matplotlib import pyplot as _plt
import pickle as _pickle
from dataclasses import dataclass as _dataclass

from .. import CarbonSimulatorUI as _CarbonSimulatorUI
from .params import Params
from .strategy import strategy as _strategy


pair_nt = namedtuple("pair", "tknb, tknq")

@_dataclass
class simresults():
    strat       : any
    spot_r      : any
    rskamt_r    : any
    cshamt_r    : any
    value_r     : any
    hodl_r      : any
    margpbuy_r  : any
    margpsell_r : any


def run_sim(strat, path, shift):
    """
    runs the simulation

    :strat:     the strategy object, or a list thereof
    :path:      the path as pandas or np series
    :shift:     the strategy shift [strat centered at (1+shift)*spot]
    :returns:   simresults tuple (rskamt_r, cshamt_r, value_r) where
                each of the ranges numpy vectors
    """
    if isinstance(strat, _strategy):
        strat = (strat,)
    slashpair = strat[0].slashpair
    for ix, strat_ in enumerate(strat):
        if not strat_.slashpair == slashpair:
            raise ValueError("All pairs must be the same", ix, slashpair, strat_.slashpair, strat[0], strat_)
    
    path_first_date =min(path.index)
    path_first_data_above_0 = min(path.index[path>0])
    if path_first_data_above_0 > path_first_date:
        print(f"[run_sim] warning: zero price data below {path_first_data_above_0} (path start {path_first_date})")
        path = path[path>0]
    
    spot = path.iloc[0]
    #print("[run_sim] rescaling", strat)
    strat = tuple(s.rescale_strat(spot*(1+shift)) for s in strat)
    
    Sim = _CarbonSimulatorUI(pair=slashpair)
    #print("[run_sim] strategies", strat)
    for strat_ in strat:
        r = Sim.add_strategy(**strat_.dct)
        #print("[run_sim] add_strategy", r)
    ouis = Sim.state()["orderuis"]

    # FACTS:
    # ouis[0].pair.tknb == ouis[0].tkn  -> buy/bid
    # ouis[1].pair.tknq == ouis[1].tkn  -> sell/ask
    # in other words
    # oui.tkn == tknb --> buy/bid
    # oui.tkn == tknq --> sell/ask
    # of course oui also has the bidask property...
    
    ouis_buy  = tuple(o for _,o in ouis.items() if o.bidask == "BID") # 0
    ouis_sell = tuple(o for _,o in ouis.items() if o.bidask == "ASK") # 1
    
    rskamt_r     = _np.array([sum(o.y for o in ouis_buy),])
    cshamt_r     = _np.array([sum(o.y for o in ouis_sell),])
    margpbuy_rz  = [tuple(o.p_marg for o in ouis_buy),]
    margpsell_rz = [tuple(o.p_marg for o in ouis_sell),]
    for spot in path[1:]:
        for oui in ouis.values():
            oui.tradeto(spot)
        rskamt_r      = _np.append(rskamt_r, [sum(o.y for o in ouis_buy),])
        cshamt_r      = _np.append(cshamt_r, [sum(o.y for o in ouis_sell),])
        
        margpbuy_rz  += [tuple(o.p_marg for o in ouis_buy),]
        margpsell_rz += [tuple(o.p_marg for o in ouis_sell),]
            # this creates margpbuy_r0 as a single series of n-tuples
            # once we are done we want to splice this into n series of singles
            # remember, if z = zip(a,b,c) then a,b,c = zip(*z)

    margpbuy_r  = tuple(zip(*margpbuy_rz))
    margpsell_r = tuple(zip(*margpsell_rz))
    value_r = rskamt_r * path + cshamt_r
    hodl_r = rskamt_r[0]*path+cshamt_r[0]
    #print(f"f[run_sim] initial amounts RSK={rskamt_r[0]}, CSH={cshamt_r[0]}", )

    return simresults(
        strat       = strat,
        spot_r      = path,
        rskamt_r    = rskamt_r, 
        cshamt_r    = cshamt_r, 
        value_r     = value_r,
        hodl_r      = hodl_r,
        margpbuy_r  = margpbuy_r,
        margpsell_r = margpsell_r
    ) 

SIM_DEFAULT_PARAMS = Params(
    plotPrice           = True,      # whether to plot the price
    plotValueCsh        = False,     # whether to plot the cash portion of the portfolio value
    plotValueRsk        = False,     # whether to plot the risk asset portion of the portfolio value
    plotValueTotal      = True,      # whether to plot the aggregate portfolio value
    plotValueHODL       = True,     # whether to plot the HODL value of the initial portfolio
    plotRanges          = True,      # whether to shade the ranges
    plotMargP           = True,      # whetger to plot the marginal price for the ranges
    plotBid             = True,      # whether to plot buy (bid) ranges and marginal prices
    plotAsk             = True,      # whether to plot sell (ask) ranges and marginal prices
    plotInterpolated    = True,      # whether to plot interpolated data
)

def plot_sim(simresults, simresults0, dataid, params, pair=None):
    """
    plots the simulation chart

    :simresults:    the simresults returned by run_sim running with path 
    :simresults0:   ditto, but for path0 (the uninterpolated path)
                    if path is path0 then simresults is simresults0
    :dataid:        a description of the data the will be used in the title
    :params:        the parameter object (can be a dict; defaults SIM_DEFAULT_PARAMS)
    :pair:          the pair as pair_nt or tuple (tknb,tknq)
    """

    has_interpolated_results = not simresults is simresults0
    strat = simresults.strat
    path = simresults.spot_r
    path0 = simresults0.spot_r
    if not pair is None: 
        pair = pair_nt(*pair)

    p = Params.construct(params, defaults=SIM_DEFAULT_PARAMS.params)
    
    if isinstance(strat, _strategy):
        strat = (strat,)
    slashpair = strat[0].slashpair
    for ix, strat_ in enumerate(strat):
        if not strat_.slashpair == slashpair:
            raise ValueError("All pairs must be the same", ix, slashpair, strat_.slashpair, strat[0], strat_)
    
    p_buy_a  = max(strat_.p_buy_a  for strat_ in strat)
    p_buy_b  = min(strat_.p_buy_b  for strat_ in strat)
    p_sell_a = min(strat_.p_sell_a for strat_ in strat)
    p_sell_b = max(strat_.p_sell_b for strat_ in strat)
    amt_rsk  = sum(strat_.amt_rsk  for strat_ in strat)
    amt_csh  = sum(strat_.amt_csh  for strat_ in strat)
    rsk      = strat[0].rsk if not pair else pair.tknb
    csh      = strat[0].csh if not pair else pair.tknq
    mid      = sqrt(p_buy_a*p_sell_a)

    descr = strat[0].descr if len(strat)==1 else f"strategy portfolio ({len(strat)} items)"
    descr = f"BID {p_buy_b:.1f}-{p_buy_a:.1f} [{amt_csh:.0f} {csh}] - MID {mid:.1f} - ASK {p_sell_a:.1f}-{p_sell_b:.1f} [{amt_rsk:.1f} {rsk}]"
    
    rskamt_r    = simresults.rskamt_r 
    cshamt_r    = simresults.cshamt_r 
    value_r     = simresults.value_r 
    value0_r    = simresults0.value_r 
    hodl_r      = simresults.hodl_r 
    margpbuy_r  = simresults.margpbuy_r 
    margpsell_r = simresults.margpsell_r

    fig, ax1 = _plt.subplots()
    ax2 = ax1.twinx()
    plots = []
    if p.plotRanges:
        for s in strat:
            if p.plotBid:
                [ax1.fill_between(path.index, s.p_buy_a, s.p_buy_b, color="lightgreen", alpha=0.1, label="bid range [lhs]")]
            if p.plotAsk:
                [ax1.fill_between(path.index, s.p_sell_a, s.p_sell_b, color="lightcoral", alpha=0.1, label="ask range [lhs]")]
            
    if p.plotMargP:
        for margpsell_ri, margpbuy_ri, ix in zip(margpsell_r, margpbuy_r, range(len(margpsell_r))):
            if p.plotBid:
                plots += ax1.plot(path.index, margpsell_ri, color="green", linestyle="dotted", linewidth=0.8, label="bid [lhs]" if not ix else None)
            if p.plotAsk:
                plots += ax1.plot(path.index, margpbuy_ri, color="red", linestyle="dotted", linewidth=0.8, label="ask [lhs]" if not ix else None)
            
    if p.plotPrice:
        plots += ax1.plot(path0, color="darkorange", alpha=0.4, label="price [lhs]")
        if has_interpolated_results and p.plotInterpolated:
            plots += ax1.plot(path, color="darkorange", alpha=0.6, linewidth=0.4)
    
    if p.plotValueHODL:
        plots += ax2.plot(value_r.index, hodl_r, color="cyan", linestyle="dotted", linewidth=1, label=f"HODL value [rhs]")
    
    if p.plotValueCsh:
        plots += ax2.plot(value_r.index, cshamt_r, color="blue", linestyle="dashed", linewidth=0.8, alpha=0.7, label=f"{csh} portion [rhs]")

    if p.plotValueRsk:
        plots += ax2.plot(value_r.index, rskamt_r*path, color="blue", linestyle="dotted", linewidth=1.2, alpha=1, label=f"{rsk} portion [rhs]")
    
    if p.plotValueTotal:
        plots += ax2.plot(value0_r, color="blue", label="portfolio value [rhs]")
        if has_interpolated_results and p.plotInterpolated:
            plots += ax2.plot(value_r, linewidth=0.2, color="royalblue", label="portfolio value [rhs]")
        
    ax2.set_ylabel(f"portfolio value ({csh})")
    ax1.set_ylabel(f"price ({csh} per {rsk})")
    #ax1.set_xlabel("date")
    _plt.title(f"{descr} on {dataid}")
    labels=[p.get_label() for p in plots]
    plots_labels=[(p,l) for l,p in zip(labels[::-1], plots[::-1]) if not l[0] == "_"]
    #_plt.legend(plots[::-1], labels[::-1])
    _plt.legend(*zip(*plots_labels))


class StartConditions():
    def __init__(self, rawdata=None):
        self._d = rawdata
        self.start_dt = _pd.Timestamp("2020-01-01")
        self.end_dt   = _pd.Timestamp("2021-01-01")
    
    @classmethod
    def load(cls, fn):
        with open(fn, 'rb') as f:
            SC=StartConditions(_pickle.load(f))
        return SC
    
    @property
    def strategy_carbon(self):
        p_sell_b, p_sell_a, p_buy_a, p_buy_b = self.carbon_ranges_raw
        psell_marginal, pbuy_marginal = self.carbon_prices
        w0, w1 = self.carbon_order_weights
        fctr = self.starting_portfolio_valuation/(w0+w1)
        amt_rsk, amt_csh = fctr*w0/self.path_raw[0], fctr*w1
        return _strategy(
            p_sell_a=p_sell_a,
            p_sell_b=p_sell_b,
            p_buy_a=p_buy_a,
            p_buy_b=p_buy_b,
            psell_marginal=psell_marginal,
            pbuy_marginal=pbuy_marginal,
            amt_rsk=amt_rsk,
            amt_csh=amt_csh,
        )
        
    @property
    def num_points(self):
        return len(self.path_raw)
    
    @property
    def time_period(self):
        return self.end_dt-self.start_dt
    
    @property
    def time_step(self):
        return self.time_period / (self.num_points-1)
    
    @property
    def path_raw(self):
        return [float(x) for x in self._d["filtered price chart"]]
    
    @property
    def path(self):
        df = _pd.DataFrame(zip(self.datetime,self.path_raw), columns=["datetime", "spot"]).set_index("datetime")
        return df["spot"]

    @property
    def datetime(self):
        dt = self.time_step
        return [self.start_dt+i*dt for i in range(self.num_points+1)]
    
    @property
    def carbon_order_weights(self):
        return [float(x) for x in self._d["carbon order weights"]]

    @property
    def carbon_ranges_raw(self):
        return [float(x) for x in self._d["carbon order boundaries"]]
    
    @property
    def carbon_prices(self):
        return [float(x) for x in self._d["carbon starting prices"]]  
   
    @property
    def starting_portfolio_valuation(self):
        return [float(x) for x in self._d["starting portfolio valuation"]][0]
    
    @property
    def uni_ranges_raw(self):
        return [float(x) for x in self._d["uniswap range boundaries"]]
    


from .fls import fload, fsave
from .params import Params
from .pdread import pdread, pathtime, pathtime_yrs, pathtime_days, pdcols, j
from .sharedvar import SharedVar
from .strategy import strategy
from .version import require_version
from .helpers import listdir
from .pathgenerator import PathGenerator
from .pathinterpolation import PathInterpolation, interpolate
from .simulation import run_sim, plot_sim, SIM_DEFAULT_PARAMS

from .. import (
    CarbonSimulatorUI as _CarbonSimulatorUI, 
    CarbonOrderUI as _CarbonOrderUI, 
    __version__, 
    __date__,
)

def print_version(require=None, all=False):
    """
    prints Carbon version numbers; calls require_version(require) if not require is None
    """
    print(f"Carbon v{__version__} ({__date__})")
    if all:
        print( "{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(_CarbonSimulatorUI))
        print( "{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(_CarbonOrderUI))

    if not require is None:
        require_version(require)

def plt_style(style, alt_style):
    """
    calls plt.style.use(style) and as fallback alt_style, all wrapped in try blocks
    """
    from matplotlib import pyplot as plt
    try:
        plt.style.use(style)
    except:
        try:
            plt.style.use(alt_style)
        except:
            print("[plt_style] both styles failed", style, alt_style)




# ------------------------------------------------------------
# Auto generated test file `test_904_MatchSlow.py`
# ------------------------------------------------------------
# source file   = NBTest_904_MatchSlow.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 904
# test comment  = MatchSlow
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))





# ------------------------------------------------------------
# Test      904
# File      test_904_MatchSlow.py
# Segment   Match
# ------------------------------------------------------------
def test_match():
# ------------------------------------------------------------
    
    # ### Set up the routers and check that the matching method has been properly accepted
    
    # First we set up a number of simulators that use the different matching methods
    
    Sim = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC")
    assert(Sim._mm==CarbonSimulatorUI.MATCH_EXACT)
    Sim
    
    SimE = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC", matching_method=CarbonSimulatorUI.MATCH_EXACT)
    assert(SimE._mm==CarbonSimulatorUI.MATCH_EXACT)
    SimE
    
    SimA = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC", matching_method=CarbonSimulatorUI.MATCH_ALPHA)
    assert(SimA._mm==CarbonSimulatorUI.MATCH_ALPHA)
    SimA
    
    #SimF = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC", matching_method=CarbonSimulatorUI.MATCH_FAST)
    SimF = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC")
    #assert(SimA._mm==CarbonSimulatorUI.MATCH_FAST)
    SimF
    
    SIMS = [Sim, SimE, SimA, SimF]
    
    # ### Set up the orders and make sure they have been accepted
    
    # We now set up a number of orders that each of the sims will get
    
    ORDERS = [
        ("ETH", 1, 2000+10*i, 3000)
        for i in range(20)   
    ]
    ORDERS[:2]
    
    for sim in SIMS:
        for o in ORDERS:
            sim.add_order(*o)
        print(sim, len(sim.state()["orders"]), len(ORDERS))
        assert(len(sim.state()["orders"])==2*len(ORDERS))
    
    # ### Run the trades and look at the routing
    
    # Now look at a trade of 1 ETH 
    
    Sim.amm_sells("ETH", 1, execute=False)["trades"]
    
    SimA.amm_sells("ETH", 1, execute=False)["trades"]
    
    


"""
Carbon helper module -- encapsulate parameters for a single strategy
"""
__VERSION__ = "2.1"
__DATE__ = "28/01/2023"


from dataclasses import dataclass as _dataclass
from math import sqrt

@_dataclass
class strategy():
    """
    class describing a Carbon strategy

    :p_sell_a:          price a (=start) of the sell/ask range*  
    :p_sell_b:          price b (=end) of the sell/ask range*  
    :p_buy_a:           price a (=start) of the buy/bid range*   
    :p_buy_b:           price b (=end) of the buy/bid range*   
    :amt_rsk:           the amount of risk asset the strategy is seeded with**
    :amt_csh:           the amount of cash the strategy is seeded with**
    :psell_marginal:    the current marginal price of the sell range
    :pbuy_marginal:     the current marginal price of the buy range
    :rsk:               risk asset name (default RSK)
    :csh:               cash name (default CSH)
    :rescale:           whether or not the strategy should be rescaled when
                        calling the rescale method
    :rescalted_fctr:    the factor this strategy was rescaled (None if not rescaled)

    *p_bid_x and p_ask_x are provided as alias properties

    **note that strategies that are seeded with 0 are instead seeded with
    `cls.MIN_SEED_AMT` which is set to 1e-10 (it can be changed in a derived
    class). The reason is that zero-seeded ranges do not allow for marginal prices.
    """
    p_sell_a: float
    p_sell_b: float 
    p_buy_a: float
    p_buy_b: float
    amt_rsk: float = 0
    amt_csh: float = 0

    psell_marginal:float = None
    pbuy_marginal:float = None

    y_int_sell:float = None
    y_int_buy:float = None
        
    rsk: str = "RSK"
    csh: str = "CSH"

    rescale: bool = True
    rescaled_fctr: float = None

    def __post_init__(self):
        pm_provided = not self.psell_marginal is None or not self.pbuy_marginal is None
        yi_provided = not self.y_int_sell is None or not self.y_int_buy is None
        if pm_provided and yi_provided:
            raise ValueError("Not both p_marginal and y_int can be provided", self.psell_marginal, self.y_int_sell, self.pbuy_marginal, self.y_int_buy)
        
    MIN_SEED_AMT = 1e-10

    def rescale_strat(self, newspot=100, oldspot=100, force=False):
        """
        rescale the strategy to another spot price (returns new object)

        :newspot:     the new spot price  to which to rescale
        :oldspot:     the old spot price basis for rescaling
        :force:       normally, strategies with rescale = False are not rescaled
                      using force = True forces a rescale
        """
        if not self.rescale:
            if not force:
                print("[rescale_strat] not rescaling", newspot, oldspot, self)
                return self
        fctr = newspot / oldspot
        #print(f"[rescale_strat] rescaling new={newspot}, old={oldspot}, fctr={fctr}", self)
        return self.__class__(
            p_sell_a = self.p_sell_a*fctr,
            p_sell_b = self.p_sell_b*fctr,
            p_buy_a  = self.p_buy_a*fctr,
            p_buy_b  = self.p_buy_b*fctr,

            amt_rsk  = self.amt_rsk,
            amt_csh  = self.amt_csh,

            psell_marginal  = self.psell_marginal*fctr if not self.psell_marginal is None else None,
            pbuy_marginal   = self.pbuy_marginal*fctr if not self.pbuy_marginal is None else None,

            y_int_sell  = self.y_int_sell,
            y_int_buy   = self.y_int_buy,

            rsk  = self.rsk,
            csh  = self.csh,

            rescale = False,
            rescaled_fctr = fctr,
        )

    def set_tvl(self, spot, cashpc, tvl):
        """
        sets amt_csh and amt_rsk based on TVL, percentage cash and spot

        :spot:      the spot price (in CSH per RSK)
        :cashpc:    the percentage of the portfolio in CSH (1.0=100%)
        :tvl:       the total portfolio value in CSH (default=1,000)
        :returns:   self for chaining
        """
        self.amt_rsk = tvl/spot*(1-cashpc)
        self.amt_csh = tvl*cashpc
        return self

    @property
    def p_bid_a(self):
        """alias for p_buy_a"""
        return self.p_buy_a
    
    @property
    def p_ask_a(self):
        """alias for p_sell_a"""
        return self.p_sell_a
    
    @property
    def p_bid_b(self):
        """alias for p_buy_b"""
        return self.p_buy_b

    @property
    def p_ask_b(self):
        """alias for p_sell_b"""
        return self.p_sell_b

    def __post_init__(self):
        if not self.amt_rsk: self.amt_risk = self.MIN_SEED_AMT
        if not self.amt_csh: self.amt_csh  = self.MIN_SEED_AMT
        if self.rsk is None: self.rsk = "RSK"
        if self.csh is None: self.csh = "CSH"

    MAX_UTIL = 0.999 # all MAXUTIL < u <= 1 are set to u = MAXUTIL

    @classmethod
    def from_mgw(cls, m=100, g=0, w=0, u=0, amt_rsk=None, amt_csh=None, rsk=None, csh=None):
        """
        create instance from mid `m`, gap width `g`, and range width `w`
        
        :m:             (geometric) middle between the two ranges
        :g:             geometric gap between the two ranges
        :w:             geometric width of the range
        :u:             percentage utilization of the range (0.01=1%; 0<=u<1)**
        :amt_rsk:       amount of risk asset the strategy is seeded with*
        :amt_csh:       amount of cash the strategy is seeded with*
        :rsk:           risk asset name (default RSK)
        :csh:           cash name (default CSH)

        *note that strategies that are seeded with 0 are instead seeded with
        `cls.MIN_SEED_AMT` which is set to 1e-10 (it can be changed in a derived
        class). The reason is that zero-seeded ranges do not allow for marginal prices.

            p_buy_a     = m/(1+g/2)
            p_buy_b     = m/(1+g/2)/(1+w)
            p_sell_a    = m*(1+g/2)
            p_sell_b    = m*(1+g/2)*(1+w)

        **a standard range has u=0, meaning that p_marg = p_start; when u->1 we have 
        p_marg -> p_end (we can not have u=1 for y>0); p_marg = (1-u) p_s + u p_e;
        for values MAX_UTIL < u <=1, u is set to MAX_UTIL
        """
        if w==0: w=0.0001
        if g==0: g=0.0001
        if u<0:
            raise ValueError("Must have u >= 0", u)
        elif u>1:
            raise ValueError("Must have u < 1", u)
        elif u > cls.MAX_UTIL:
            u = cls.MAX_UTIL
        
        g2 = 0.5*g
        p_buy_a     = m/(1+g2)
        p_buy_b     = m/(1+g2)/(1+w)
        p_sell_a    = m*(1+g2)
        p_sell_b    = m*(1+g2)*(1+w)

        return cls(
            p_buy_a = p_buy_a, 
            p_buy_b = p_buy_b,
            p_sell_a = p_sell_a,
            p_sell_b = p_sell_b,
            amt_rsk = amt_rsk,
            amt_csh = amt_csh,
            rsk = rsk,
            csh = csh,
            psell_marginal = (1-u)*p_sell_a + u*p_sell_b if u else None,
            pbuy_marginal = (1-u)*p_buy_a + u*p_buy_b if u else None,
        )
    
    @classmethod
    def from_u3(cls, p_lo, p_hi, start_below, tvl_csh=None, fee_pc=None, rsk=None, csh=None):
        """
        creates a strategy equivalent to a uniswap v3 position

        :p_lo:          the low price of the range
        :p_hi:          the high price of the range
        :fee_pc:        percentage fees (0.1=1%)
        :start_below:   if True/False, the start price is below/above the range, so 
                        the combined range is 100% in RSK/CSH p_marginal at p_lo/p_hi
        :tvl_csh:       the TVL (rsk+csh) at top of the range, measured in csh (default: 1000)
        :rsk:           risk asset name (default RSK)
        :csh:           cash name (default CSH)
        """
        #assert fee_pc is None, "fees not implemented yet"
        if not p_hi >= p_lo:
            raise ValueError("Must have p_hi > p_lo", p_hi, p_lo)
        if tvl_csh is None:
            tvl_csh = 1000

        if start_below:
            p_sell_a = p_buy_b = p_lo   
            p_sell_b = p_buy_a = p_hi   
        else:
            p_sell_a = p_buy_b = p_hi  
            p_sell_b = p_buy_a = p_lo   

        if not fee_pc is None:
            #fee_mult = 1+0.5*fee_pc
            fee_shift = sqrt(p_lo*p_hi)*fee_pc*0.5
            # p_sell_a *= fee_mult
            # p_sell_b *= fee_mult
            # p_buy_a  /= fee_mult
            # p_buy_b  /= fee_mult
            p_sell_a += fee_shift
            p_sell_b += fee_shift
            p_buy_a  -= fee_shift
            p_buy_b  -= fee_shift

        if start_below:
            #p_marginal = p_lo*1.0000000001
            amt_rsk = 100 # tvl_csh/sqrt(p_lo*p_hi)
            amt_csh = 0
        else:
            #p_marginal = p_hi*0.9999999999
            amt_rsk = 0
            amt_csh = tvl_csh

        return cls(
            p_buy_a = p_buy_a, 
            p_buy_b = p_buy_b,
            p_sell_a = p_sell_a,
            p_sell_b = p_sell_b,
            amt_rsk = amt_rsk,
            amt_csh = amt_csh,
            rsk = rsk,
            csh = csh,
            # psell_marginal = p_marginal*fee_mult,
            # pbuy_marginal = p_marginal/fee_mult,
            y_int_sell = 100, # RSK NUMBER
            y_int_buy = 100 * sqrt(p_lo*p_hi), # CSH NUMBER
        )

    @property
    def descr(s):
        """returns a description of the strategy"""
        bid_s = f"BID {s.p_buy_b:.1f}-{s.p_buy_a:.1f} [{s.amt_csh} {s.csh}]"
        ask_s = f"ASK {s.p_sell_a:.1f}-{s.p_sell_b:.1f} [{s.amt_rsk} {s.rsk}]"
        mid_s = f"MID {sqrt(s.p_buy_a*s.p_sell_a):.1f}"
        return f"{bid_s} - {mid_s} - {ask_s}"
    
    @property
    def p(s):
        """DEPRECTATED returns a tuple suitable for `add_strategy (*params)`"""
        print("[strategy::p] DEPRECATED. PLEASE USE dct INSTEAD")
        return (s.rsk, s.amt_rsk, s.p_sell_a, s.p_sell_b, s.amt_csh, s.p_buy_a, s.p_buy_b)
    
    @property
    def dct(s):
        """returns a dict suitable for `add_strategy (**params)`"""
        return {
            "tkn":              s.rsk, 
            "amt_sell":         s.amt_rsk, 
            "psell_start":      s.p_sell_a, 
            "psell_end":        s.p_sell_b, 
            "amt_buy":          s.amt_csh, 
            "pbuy_start":       s.p_buy_a, 
            "pbuy_end":         s.p_buy_b,
            "pair":             s.slashpair,
            "psell_marginal":   s.psell_marginal,
            "pbuy_marginal":    s.pbuy_marginal,
            "y_int_sell":       s.y_int_sell,
            "y_int_buy":        s.y_int_buy,
        }

        # def add_strategy(
        #     self,
        #     tkn: str,
        #     amt_sell: Any = None,
        #     psell_start: Any = None,
        #     psell_end: Any = None,
        #     amt_buy: Any = None,
        #     pbuy_start: Any = None,
        #     pbuy_end: Any = None,
        #     pair: str = None,
        #     psell_marginal: Any = None,
        #     pbuy_marginal: Any = None,
        #     y_int_sell: Any = None,
        #     y_int_buy: Any = None, 
        # ) -> Dict[str, Any]:
    
    @property
    def slashpair(self):
        """returns the slashpair as str, eg 'RSK/CSH'"""
        return f"{self.rsk}/{self.csh}"



# ------------------------------------------------------------
# Auto generated test file `test_052_Demo73Helpers.py`
# ------------------------------------------------------------
# source file   = NBTest_052_Demo73Helpers.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 052
# test comment  = Demo73Helpers
# ------------------------------------------------------------



from carbon.helpers.stdimports import *

from carbon.helpers.version import VersionRequirementNotMetError
from carbon.helpers.simulation import run_sim, plot_sim
from carbon.helpers.widgets import CheckboxManager, DropdownManager, PcSliderManager

from carbon.helpers import fload, fsave
from carbon.helpers import Params
from carbon.helpers import pdread, pathtime, pathtime_yrs, pathtime_days, pdcols, j
from carbon.helpers import strategy
from carbon.helpers import require_version
from carbon.helpers import print_version

plt.style.use('seaborn-dark')
plt.rcParams['figure.figsize'] = [12,6]
print_version(require="2.2.4")




# ------------------------------------------------------------
# Test      052
# File      test_052_Demo73Helpers.py
# Segment   params
# ------------------------------------------------------------
def test_params():
# ------------------------------------------------------------
    
    p = Params(a=1, b=2)
    assert p["a"] == 1
    assert p.a == 1
    assert p["c"] is None
    print(p)
    assert str(p) == "Params.construct({'a': 1, 'b': 2})"
    assert p.params == {'a': 1, 'b': 2}
    p["c"] = 5
    assert p["c"] == 5
    assert p.c == 5
    result = p.add(d=10, e=11)
    assert result is p
    assert p["d"] == 10
    assert p.e == 11
    try:
        p.z
        raise RuntimeError("should raise")
    except KeyError as e:
        print(e)
    
    # +
    p = Params(a=1, b=2)
    assert p.defaults == {}
    p.set_default(b=20, c=3)
    assert str(p) == "Params.construct({'a': 1, 'b': 2}, defaults={'b': 20, 'c': 3})"
    assert p.b == 2
    assert p.get_default("b") == 20
    assert p.c == 3
    
    p = Params(a=1, b=2)
    assert p.get_default("c") is None
    result = p.set_default()
    assert result is p
    p.set_default(c=10, d=11)
    assert p.c == 10
    assert p.d == 11
    assert p.defaults == {'c': 10, 'd': 11}
    assert p["e"] is None
    try:
        p.e
        raise RuntimeError("should raise")
    except KeyError as e:
        pass
    # -
    
    p = Params.construct({'a': 1, 'b': 2}, defaults={'b': 20, 'c': 3})
    assert p.a == 1
    assert p.b == 2
    assert p.c == 3
    assert p.get_default("b") == 20
    pp = Params.construct(p)
    assert not pp is p
    assert p.params == pp.params
    try:
        ppp = Params.construct(p, defaults={"e":100})
    except ValueError as e:
        print(e)
    
    p  = Params.construct({'a': 1, 'b': 2})
    defaults = {'b': 20, 'c': 3}
    pp = Params.construct(p, defaults=defaults)
    assert not pp is p
    assert p.params == pp.params
    assert pp.defaults == defaults
    assert not pp.defaults is defaults
    assert not pp._defaults is defaults
    
    p = Params.construct()
    assert p.params == {}
    p = Params.construct(None)
    assert p.params == {}
    

# ------------------------------------------------------------
# Test      052
# File      test_052_Demo73Helpers.py
# Segment   widgets [NOTEST]
# ------------------------------------------------------------
def notest_widgets():
# ------------------------------------------------------------
    #
    # https://ipywidgets.readthedocs.io/en/7.6.3/examples/Widget%20Styling.html#The-Flexbox-layout
    
    CM1 = CheckboxManager([i for i in range(10)])
    CM1(vertical=False)
    
    CM2 = CheckboxManager({i:678975437*(i+10) for i in range(5)}, objects=("a", "b", "c", "d", "e"))
    CM2()
    
    CM2.objects(), CM2.objects(checked=False)
    
    SM = PcSliderManager([f"Slider #{i+1}" for i in range(5)])
    SM()
    
    SM.values
    
    DD = DropdownManager([678975437*(i+10) for i in range(5)])
    DD()
    
    print(CM1.values, CM2.values, DD.value)
    

# ------------------------------------------------------------
# Test      052
# File      test_052_Demo73Helpers.py
# Segment   widgets
# ------------------------------------------------------------
def test_widgets():
# ------------------------------------------------------------
    
    CM = CheckboxManager([1,2,3], None)
    assert CM.choice_ids == ('1', '2', '3')
    assert CM.choice_descr == ('1', '2', '3')
    assert CM.values == (False, False, False)
    assert CheckboxManager({1:11, 2:12}).choice_ids == ('1', '2')
    assert CheckboxManager({1:11, 2:12}).choice_descr == ('11', '12')
    assert CheckboxManager([1,2,3], False).values == (False, False, False)
    assert CheckboxManager([1,2,3], True).values == (True, True, True)
    assert CheckboxManager([1,2,3], [10,0,1]).values == (True, False, True)
    assert CheckboxManager([1,2,3], disabled=[10,0,1]).disabled == (True, False, True)
    assert CheckboxManager.from_idvdct({10:10,12:0}).choice_ids == ('10', '12')
    assert CheckboxManager.from_idvdct({10:10,12:0}).choice_descr == ('10', '12')
    assert CheckboxManager.from_idvdct({10:10,12:0}).values == (True, False)
    try:
        CheckboxManager([1,2,3], [10,0])
        raise
    except ValueError as e:
        print(e)
    try:
        CheckboxManager([1,2,3], disabled=[10,0])
        raise
    except ValueError as e:
        print(e)
    
    assert PcSliderManager({1:11, 2:12}).choice_ids == ('1', '2')
    assert PcSliderManager({1:11, 2:12}).choice_descr == ('11', '12')
    assert PcSliderManager([1,2,3], None).values == (0.,)*3
    assert PcSliderManager([1,2,3], 0.3).values == (0.3, 0.3, 0.3)
    assert PcSliderManager([1,2,3], [2,0.5,-1]).values == (1, 0.5, 0)
    assert PcSliderManager([1,2,3], disabled=[10,0,1]).disabled == (True, False, True)
    try:
        PcSliderManager([1,2,3], [10,0])
        raise
    except ValueError as e:
        print(e)
    try:
        PcSliderManager([1,2,3], disabled=[10,0])
        raise
    except ValueError as e:
        print(e)
    
    assert DropdownManager([1,2,3]).options == ('1', '2', '3')
    assert DropdownManager([1,2,3]).labels == ('1', '2', '3')
    assert DropdownManager([1,2,3], labels=[10,20,30]).options == ('1', '2', '3')
    assert DropdownManager([1,2,3], labels=[10,20,30]).labels == ('10', '20', '30')
    assert DropdownManager({1:10, 2:20, 3:30}).options == ('1', '2', '3')
    assert DropdownManager({1:10, 2:20, 3:30}).labels == ('10', '20', '30')
    assert DropdownManager({1:10, 2:20, 3:30}).value == "1"
    assert DropdownManager({1:10, 2:20, 3:30}).label == "10"
    assert DropdownManager({1:10, 2:20, 3:30}).result == ("1", "10")
    

# ------------------------------------------------------------
# Test      052
# File      test_052_Demo73Helpers.py
# Segment   helpers stdimport
# ------------------------------------------------------------
def test_helpers_stdimport():
# ------------------------------------------------------------
    
    # check that the following objects have been imported
    
    np
    plt
    pd
    CarbonSimulatorUI
    CarbonOrderUI
    P
    
    sqrt
    exp
    log
    

# ------------------------------------------------------------
# Test      052
# File      test_052_Demo73Helpers.py
# Segment   helpers version
# ------------------------------------------------------------
def test_helpers_version():
# ------------------------------------------------------------
    
    assert require_version("1.0", "1.0", raiseonfail=False) == True
    assert require_version("2.0", "1.0", raiseonfail=False) == False
    assert require_version("1.0", "2.0", raiseonfail=False) == True
    assert require_version("1.0.1", "1.0", raiseonfail=False) == False
    assert require_version("1.0", "1.0.1", raiseonfail=False) == True
    assert require_version("1.0", "11.0", raiseonfail=False) == True
    assert require_version("11.0", "1.0", raiseonfail=False) == False
    assert require_version("1.3beta1", "1.3", raiseonfail=False) == True
    assert require_version("1.3beta1", "1.3beta1", raiseonfail=False) == True
    assert require_version("1.3-1", "1.3-2", raiseonfail=False) == True
    assert require_version("1.3-2", "1.3-1", raiseonfail=False) == True
    assert require_version("1.0", "1.0", raiseonfail=False) == True
    assert require_version("1.0", "1.0", raiseonfail=False) == True
    
    # obviously must be tested with version >= 2.0
    
    require_version("2.0")
    
    require_version("1.0", "1.0")
    
    try:
        require_version("2.0", "1.0")
        run("must raise error")
    except VersionRequirementNotMetError as e:
        print(e)
    

# ------------------------------------------------------------
# Test      052
# File      test_052_Demo73Helpers.py
# Segment   helpers strategy
# ------------------------------------------------------------
def test_helpers_strategy():
# ------------------------------------------------------------
    
    # +
    # from carbon.helpers import strategy as _strategy
    # help(_strategy)
    
    # +
    # assert strategy.from_mgw() == strategy(p_buy_a=100.0, p_buy_b=100.0, p_sell_a=100, p_sell_b=100, 
    #                                        amt_rsk=0, amt_csh=0, rsk='RSK', csh='CSH')
    # -
    
    assert strategy.from_mgw(m=100) == strategy.from_mgw()
    assert strategy.from_mgw(g=0) == strategy.from_mgw()
    assert strategy.from_mgw(w=0) == strategy.from_mgw()
    
    assert strategy.from_mgw(g=0.1).p_buy_a == 100/(1.05)
    assert strategy.from_mgw(g=0.1).p_sell_a == 100*(1.05)
    assert strategy.from_mgw(g=0.1).p_buy_b/strategy.from_mgw(g=0.1).p_buy_a > 0.999
    assert strategy.from_mgw(g=0.1).p_buy_b/strategy.from_mgw(g=0.1).p_buy_a < 1.001
    assert strategy.from_mgw(g=0.1).p_sell_b / strategy.from_mgw(g=0.1).p_sell_a < 1.001
    assert strategy.from_mgw(g=0.1).p_sell_b / strategy.from_mgw(g=0.1).p_sell_a > 0.999
    
    assert strategy.from_mgw(w=0.1).p_buy_a / strategy.from_mgw(w=0.1).p_sell_a > 0.999
    assert strategy.from_mgw(w=0.1).p_buy_b / (100/1.1) > 0.999
    assert strategy.from_mgw(w=0.1).p_buy_b / (100/1.1) < 1.001
    assert strategy.from_mgw(w=0.1).p_sell_b / (100*1.1) > 0.999
    assert strategy.from_mgw(w=0.1).p_sell_b / (100*1.1) < 1.001
    
    assert strategy.from_mgw().slashpair == "RSK/CSH"
    assert strategy.from_mgw(rsk="ETH", csh="USD").slashpair == "ETH/USD"
    
    assert strategy(90, 80, 110, 120).dct == {
        'tkn': 'RSK',
        'amt_sell': 0,
        'psell_start': 90,
        'psell_end': 80,
        'amt_buy': 1e-10,
        'pbuy_start': 110,
        'pbuy_end': 120,
        'pair': 'RSK/CSH',
        'psell_marginal': None,
        'pbuy_marginal': None,
        'y_int_sell': None,
         'y_int_buy': None
    }
    
    assert strategy(90, 80, 110, 120).p == ('RSK', 0, 90, 80, 1e-10, 110, 120)
    
    Sim = CarbonSimulatorUI(pair="RSK/CSH", raiseonerror=True)
    r = Sim.add_strategy(**strategy(p_buy_a=80, p_buy_b=70, p_sell_a=110, p_sell_b=120, amt_rsk=1, amt_csh=100).dct)["orders"]
    r
    
    assert r.iloc[0]["p_start"] == 110
    assert r.iloc[0]["p_end"] == 120
    assert r.iloc[1]["p_start"] == 80
    assert r.iloc[1]["p_end"] == 70
    assert r.iloc[0]["y"] == 1
    assert r.iloc[1]["y"] == 100
    
    Sim = CarbonSimulatorUI(pair="RSK/CSH", raiseonerror=True)
    r = Sim.add_strategy(**strategy.from_mgw(m=100, g=0.5, w=0.5, amt_rsk=1, amt_csh=100).dct)["orders"]
    r
    
    assert r.iloc[0]["p_start"] == 125
    assert r.iloc[0]["p_end"] == 187.5
    assert r.iloc[1]["p_start"] == 80
    assert r.iloc[0]["y"] == 1
    assert r.iloc[1]["y"] == 100
    

# ------------------------------------------------------------
# Test      052
# File      test_052_Demo73Helpers.py
# Segment   helpers pdread
# ------------------------------------------------------------
def test_helpers_pdread():
# ------------------------------------------------------------
    
    # +
    # from carbon.helpers import pdread as _pdread
    # help(_pdread)
    # -
    
    DATAFN = "resources/data/RAN-050-00.pickle"
    #DATAFN = "../data/RAN-050-00.pickle"
    
    assert len(pdread(DATAFN))==101
    assert pdread(DATAFN).iloc[0].index[5] == "p0005"
    assert str(pdread(DATAFN).index[0]) == '2020-01-01 00:00:00'
    assert len(pdread(DATAFN)["p0000"]) == len(pdread(DATAFN))
    assert list(pdread(DATAFN)["p0000"]) == list(pdread(DATAFN, "p0000"))
    
    # those assertions should be (mostly) independent from the test data used
    ONEDAY = pd.Timedelta(days=1)
    FDAY, LDAY = pdread(DATAFN).index[0], pdread(DATAFN).index[-1]
    NDAY = LDAY - FDAY
    assert pdread(DATAFN, from_ts=FDAY - 100*ONEDAY).index[-1] == LDAY
    assert pdread(DATAFN, from_ts=LDAY - 100*ONEDAY).index[-1] == LDAY
    assert pdread(DATAFN, from_ts=FDAY + 100*ONEDAY).index[-1] == LDAY
    assert len(pdread(DATAFN, from_ts=LDAY + 1*ONEDAY))==0
    assert pdread(DATAFN, from_ts=FDAY+20*ONEDAY).index[0] >= FDAY
    assert pdread(DATAFN, period_days=NDAY/ONEDAY+10).index[-1] == LDAY
    assert pdread(DATAFN, period_days=NDAY/ONEDAY-10).index[-1] <= LDAY
    assert len(pdread(DATAFN, period_days=0))==1
    
    # those assertions depend on the exact test data used
    assert str(pdread(DATAFN                ).index[-1]) == '2020-12-31 06:00:00'
    assert str(pdread(DATAFN, period_days=20).index[-1]) == '2020-01-19 06:18:00'
    assert str(pdread(DATAFN, period_days=50).index[-1]) == '2020-02-17 11:34:48'
    
    # those assertions depend on the exact test data used
    assert str(pdread(DATAFN, from_ts="2000-01-01").index[0])  == '2020-01-01 00:00:00'
    assert str(pdread(DATAFN, from_ts="2020-01-02").index[0])  == '2020-01-04 15:39:36'
    assert str(pdread(DATAFN, from_ts="2020-01-02").index[-1]) == '2020-12-31 06:00:00'
    assert str(pdread(DATAFN, from_ts="2020-05-02").index[-1]) == '2020-12-31 06:00:00'
    
    # those assertions depend on the exact test data used
    assert str(pdread(DATAFN, from_ts="2000-01-01", period_days=30).index[0])   == '2020-01-01 00:00:00'
    assert str(pdread(DATAFN, from_ts="2000-01-01", period_days=30).index[-1])  == '2020-01-30 05:16:48'
    assert str(pdread(DATAFN, from_ts="2020-03-01", period_days=30).index[0])   == '2020-03-03 02:13:12'
    assert str(pdread(DATAFN, from_ts="2020-03-01", period_days=30).index[-1])  == '2020-03-28 15:50:24'
    
    assert pathtime(pdread(DATAFN, "p0000")) == 1       # DEPRECATED
    assert pathtime_yrs(pdread(DATAFN, "p0000")) == 1
    assert pathtime_days(pdread(DATAFN, "p0000")) == 365.25
    
    assert len(pdcols(DATAFN)) == 1000
    assert (pdcols(DATAFN) == pdread(DATAFN).columns).all() == True
    assert len(pdread(DATAFN, period_days=10)) == 3
    assert len(pdread(DATAFN, period_pc=0.1)) == 11
    assert len(pdread(DATAFN, from_ts="2020-07-01")) == 51
    assert len(pdread(DATAFN, from_pc=0.5)) == 51
    

# ------------------------------------------------------------
# Test      052
# File      test_052_Demo73Helpers.py
# Segment   demo 7 3 [NOTEST]
# ------------------------------------------------------------
def notest_demo_7_3():
# ------------------------------------------------------------
    
    # +
    DATAID = "RAN-050-00"
    
    DATAPATH = "resources/data"
    #DATAPATH = "../data"           # uncomment to run this as Jupyter notebook
    
    DATAFN = j(DATAPATH, f"{DATAID}.pickle")
    print(f"Chose data id {DATAID}")
    # -
    
    strats = (
        strategy.from_mgw(m=100, g=0.10, w=0.05, amt_rsk=1, amt_csh=0),
    )
    
    for colnm in ["p0000"]:
        for ix, strat in enumerate(strats):
        
            path = pdread(DATAFN, colnm)
            simresults = run_sim(strat, path)
            if DATAPATH == "../data":
                plot_sim(strat, path, simresults, f"{DATAID} [{colnm}]", params=None)
                plt.show()
    
    
    # +
    # assert simresults.rskamt_r[0] == 1
    # assert round(simresults.rskamt_r[35] - 0.57736478, 5) == 0
    # assert simresults.cshamt_r[0] == 0
    # assert round(simresults.cshamt_r[36] - 112.7164584255556, 5) == 0
    # assert round(simresults.value_r[4] - 98.135103, 5) == 0
    # assert round(simresults.value_r[-1] - 109.973544, 5) == 0
    # assert str(simresults.value_r.index[0]) == '2020-01-01 00:00:00'
    # assert str(simresults.value_r.index[-1]) == '2020-12-31 06:00:00'
    # -
    
    simresults.rskamt_r[35]
    
    